#!/usr/local/bin/perl
########################################################################
# PROGRAM: sentiment.pl 
#
# Version 1.0
#
# Copyright (C), 2002
# This library is free software; you can redistribute it and/or modify
# it under the same terms as Perl itself.
#
#          ***  UNION MINAS  ***
# Bridget McInnes,       bthomson@d.umn.edu
# Deodatta Y Bhoite,     bhoi0001@d.umn.edu
# Yanhua Li,             lixx0333@d.umn.edu
# Nitin Agarwal,         agar0054@d.umn.edu
# Kailash Aurangabadkar, aura0011@d.umn.edu
#
# University of Minnesota, Duluth
#
# USAGE:
#          perl sentiment.pl directory
#
# This program determines the classification of a review, ie 
# whether it is a positive review, a negative review or undecided. 
# The classification is determined through three sources. Longman's 
# Machine Readable Dictionary, Macquarie's Machine Readable Dictionary 
# and the World Wide Web. These resources are accessed respectively 
# through the following Perl Modules: 
#
#      1. Minas::LDOCE   
#      2. Minas::BigMac   
#      3. Minas::Oyster  
#  
#
# The program inputs a directory given by the user. The directory 
# contains two directories, 'neg' and 'pos'. The program reads the
# files (reviews) from these directories and processes them. The 
# results of the classification is displayed in the following way:
#
#      filename   known classification   system classification
# 
# Lastly the program then calculates the precision, recall and the 
# f-measure of the total results of the program.
#
# A configuration file is distributed with this program called
# sentiment.config. Please read the README distributed with this 
# program to modify this configurable parameters for this program.
#
# The description of the algorithm implemented in this program
# can be found in the algorithm.txt distributed with this program
#
#
##########################################################################

#use modules
use LDOCE::LDOCE;
use BigMac::BigMac;
use WebReader::WebReader;

#please note that the names of the modules are different because they
#are not installed on the system as of yet.

#get command line variable
my $directory = shift @ARGV;
chomp $directory;

#set modules
print "Initializing LDOCE....\n";
my $ldoce = Minas::LDOCE->new("/home/cs5/bthomson/CS8761/LDOCE3/A-Z/sgml/");
print "LDOCE initialized\n";
print "Initializing BigMac....\n";
my $bm = Minas::BigMac->new;
$bm->setDictionaryLocation("/home/cs/tpederse/CS8761/BigMac/bigmac.corrected2.sgml");
$bm->setThesaurusLocation("/home/cs/tpederse/CS8761/BigMac/thes.data");
print "Bigmac initialized\n";
print "Initializing WebReader....\n";
my $oyster = Minas::WebReader->new;
print "WebReader initialized\n";

#reading the configuration file and storing name value pairs in a hash table
open(CONFIG,"sentiment.config");
while($line=<CONFIG>){
    chomp $line;
    unless($line=~m/^\#/){
	($name,$value)=split(/=/,$line);
	$confighash{$name}=$value;
    }
}

#read trace file name from configuration file
$tracefile = $confighash{"tracefile"};
#open trace file
open(TRACE, ">$tracefile");

foreach $key (keys %confighash){
    print "$key = $confighash{$key}\n";
    print TRACE "$key = $confighash{$key}\n";
}

#set active hash. read active code classification
my %activ = ();
open(ACTIV, "activecodes.txt");
while(<ACTIV>) {
    chomp;
    my @codes = split/\s+/, $_;
    $s = shift @codes;
    $c = join " ", @codes;
    $activ{$c} = $s;
}
close(ACTIV);

#set adjective hash. read adjective classification
my %adj = ();
open(ADJ, "adj.txt");
while(<ADJ>) {
    chomp;
    my @codes = split /\s+/, $_;
    $adj{$codes[0]} = $codes[1];
}
close(ADJ);

#set thesaurus hash. read thesaurus classification
my %keytags = ();
open(KEYS,"theskeys.txt");
while($line=<KEYS>){
    chomp($line);
    @cols = split(/\t/,$line);
    @dots = split(/\./,$cols[0]);
    if($cols[2] eq "+"){
	$keytags{$dots[0]}=1;
    }elsif($cols[2] eq "-"){
	$keytags{$dots[0]}=-1;
    }elsif($cols[2] eq "0"){
	$keytags{$dots[0]}=0;
    }
}
close(KEYS);

#read good adjectives to find words at distance one from them
#origin words for good list
open(DG,"distgood.txt") or die "Cannot open distgood.txt file\n";
while($line = <DG>){
    chomp($line);
    $adj1{$line}="GOOD";
}
close(DG);

#read bad adjectives to find words at distance one from them
#origin words for bad list
open(DB,"distbad.txt") or die "Cannot open distbad.txt file\n";
while($line=<DB>){
    chomp($line);
    $adj1{$line}="BAD";
}
close(DB);

#set stoplist
my %stoplist = ();
open(LIST, "stoplist") or die "Cannot open stoplist file\n";
while(<LIST>) {
    chomp;
    $stoplist{$_}++;
}
close(LIST);

print "Building distance vectors for BigMac...\n";
# find words at the distance one from good and bad words
foreach $word (%adj1){
    if($adj1{$word} eq "GOOD"){
	%wordsindef= $bm->getWords($word);
	foreach $str (%wordsindef){
	    @arr = split(/;/,$str);
	    if($ldoce->isadjective($arr[0])){
		unless(exists $stoplist{$arr[0]}){
		    $goodhash{$arr[0]}=1;
		}
	    }
	}
	$goodhash{$word}=1;
    }else{
	%wordsindef= $bm->getWords($word);
	foreach $str (%wordsindef){
	    @arr = split(/;/,$str);
	    if($ldoce->isadjective($arr[0])){
	        unless(exists $stoplist{$arr[0]}){
		    $badhash{$arr[0]}=1;
		}
	    }
	}
	$badhash{$word}=1;
   }
}
print "Distance vectors built\n";

#get horrible and excellent counts for measure
$goodword = $confighash{"goodword"};
$badword = $confighash{"badword"};
my $excellentcount = $oyster->getquerynumber($goodword);          
my $horriblecount  = $oyster->getquerynumber($badword);  
#assumption of npp is total hits for excellent plus total hist for horrible
my $totalcount = $excellentcount + $horriblecount;

#set cache's to speed up program
my %activCache = ();    
my %reviewCacheWords = ();
my %goodDiceReviewCache = ();
my %badDiceReviewCache = ();
my %goodPoissonReviewCache = ();
my %badPoissonReviewCache = ();
my %goodTscoreReviewCache = ();
my %badTscoreReviewCache = ();

#set directories to open
my $neg = $directory . 'neg';
my $pos = $directory . 'pos';

my @files = ();
my @subdirectories = ();

push @subdirectories, $neg;
push @subdirectories, $pos;

#open directory to get file names
foreach my $dir (@subdirectories) {
    opendir(DIR, $dir) or die "can not open directory: $dir \n";
    #line below comes from a mix on page 770 in Programming in Perl
    push @files, grep { -T } map {"$dir/$_"} readdir(DIR); 
    close(DIR);
}

#set format for standard out
format STDOUT = 
@<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<<  @<<<<<<<<< @<<<<<<<<<<
$file, $known, $results
.



#open and classify review
my $review = "";
my $correct = 0;
my $wrong = 0;
my $total = 0;
foreach $file (@files) {
    #get the review
    open(REV, $file) or die "can not open file: $file \n";
    $review = "";
    while(<REV>) { $review = $review . $_; }              #get review
    close(REV);

    #determine possible classifications
    print TRACE "$file \n";                                #print file to trace
    print TRACE " active results:  ";
    my $active = activ();                                 #determine activ classification
    print TRACE "  => $active\n";                         #print results to trace
    print TRACE " adj results:     ";
    my $adjective = adjectives();                         #determine adj classification
    print TRACE "  => $adjective\n";                      #print results to trace
    #print TRACE " synonym results: ";               
    #my $synonym = synonymTest();                          #determine synonym classification
    #print TRACE "  => $synonym\n";                        #print results to trace
    print TRACE " measures results: ";
    my $measure = measures();                               #determine dice classification
    print TRACE " => $measure\n";                          #print results to trace
    print TRACE " distance results: ";             
    my $distance = distance();
    print TRACE " => $distance \n";

    #determine final classification by using weights from the configuration file
    my $classification = ($confighash{"activeweight"} * $active) + 
	                 ($confighash{"adjectiveweight"} * $adjective) + 
			 ($confighash{"measureweight"} * $measure) + 
			 ($confighash{"distanceweight"} * $distance);     
 
    #determine results based on classification
    $known = "";
    if($file=~m/neg/) {
	$known = "negative";                              #set known
	$classification < 0 ? $correct++ : $wrong++;      #check if neg correct
	if($classification == 0) { $unknown++; }
    }
    else {
	$known = "positive";                              #set known
	$classification > 0 ? $correct++ : $wrong++;      #check if pos correct
	if($classification == 0) { $unknown++; }
    }
    
    #determine results based on classification
    $results = $classification < 0 ? "negative" : ($classification == 0 ? "unknown" : "positive");

    #print final results to trace
    print TRACE " final results => $results\n";

    #print file known and results to standard out
    write;
} 

#determine number wrong
$wrong = $wrong - $unknown;
print " $correct  $wrong  $unknown\n";
#calculate precision, recall and fmeasure
my $precision = $correct/($correct + $wrong);
my $recall = $correct/($correct + $wrong + $unknown);
my $fmeasure = (2 * $precision * $recall) / ($precision + $recall);

#display precision, recall and fmeasure to user
print "Precision  => $precision\n";
print "Recall     => $recall\n";
print "F-measure  => $fmeasure\n";

#determines the classification of a review using the active codes 
#from the Longmans Machine Readable Dictionary  
sub activ
{  
    #initialize variables
    my $good = 0;
    my $bad = 0;
    
    #clean up reviews
    $review=~s/[.!?\"\';:%&{}(),]//g;                                      

    #determine possible classifications
    my @reviews = split/\s+/, $review;                                       
    foreach my $word (@reviews) { 
	if(exists $stoplist{$word}) {}
	elsif(exists $activCache{$word}) {
	    @activiatingarray = $activCache{$word};
	    foreach $a (@activatingarray){ 
		$a eq "GOOD" ? $good++ : $bad++;  
	    }  
	} 
	else {
	    my %code =  $ldoce->getactivecode($word); 
	    @array = ();
	    foreach my $c (sort keys %code) { 
		$activ{$c} eq "GOOD" ? $good++ : $bad++ if exists $activ{$c};
		$activ{$c} eq "GOOD" ? push @array, "GOOD" : push @array, "BAD";
	    }   
	    %activCache = [ @array ];
	} 
    }

    #print results to trace
    print TRACE "GOOD: $good BAD: $bad  ";

    #return -1 if bad, 1 if good and 0 if don't know
    ($good == $bad) ? (return 0) : ( ($good < $bad) ? (return -1) : (return 1) );
} 

#determines the classification of a review using the adjectives from the
#Longman's Machine Readable Dictionary
sub adjectives
{ 
    #initialize variables
    my $good = 0;
    my $bad = 0;

    #clean up reviews
    $review=~s/[.!?\"\';:%&{}(),]//g;                                            #remove punctuation

    #determine possible classifications
    my @reviews = split/\s+/, $review;                                           #split on words
    my $presedingword = "";
    foreach my $word (@reviews) {
	if(exists $stoplist{$word}) {}                                           #stoplist
        elsif($presedingword eq "not") { 
	    $adj{$word} eq "GOOD" ? $bad++ : $good++ if exists $adj{$word};
	}
	else { 
	    $adj{$word} eq "GOOD" ? $good++ : $bad++ if exists $adj{$word}; 
	}
	$presedingword = $word;
    }
    
    #print results to trace
    print TRACE "GOOD: $good BAD: $bad  ";

    #return -1 if bad, 1 if good and 0 if don't know
    ($good == $bad) ? (return 0) : ( ($good < $bad) ? (return -1) : (return 1) );
}

#determines the classification of the review using the various measures of association
#applied to numbers of hits for phrases gathered using the WebReader module.
sub measures
{    
    my $dicegood = 0;
    my $dicebad = 0;
    my $poissongood = 0;
    my $poissonbad = 0;
    my $tscoregood = 0;
    my $tscorebad = 0;
   
    #clean up reviews
    $review=~s/[.!?\"\';:%&{}(),]//g;                                   #remove punctuation

    #determine dice coefficient, Poisson measure and T-score of review
    my @reviews = split/\s+/, $review;                                  #split on words
    foreach my $word (@reviews) {
	if(exists $adj{$word}) { #$ldoce->isadjective($word) == 1) { 
	    if(exists $reviewCacheWord{$word}) {
		#calculate dice
		$dicegood += $goodDiceReviewCache{$word};
		$dicebad += $badDiceReviewCache{$word};
		#calculate poisson
		$poissongood += $goodPoissonReviewCache{$word};
		$poissonbad += $badPoissonReviewCache{$word};
		#calculate t score
		$tscoregood += $goodTscoreReviewCache{$word};
		$tscorebad += $badTscoreReviewCache{$word};
	    }
	    else {
		$reviewCacheWords{$word}++;                             #set word in cache

		my $horrible = $badword . "+" . $word;                  #set horrible query
		my $excellent = $goodword . "+" . $word;                #set excellent query
		$reviewcount     = $oyster->getquerynumber($word);      #get wrd count
		$reviewhorrible  = $oyster->getquerynumber($horrible);  #get horrible+wrd count
		$reviewexcellent = $oyster->getquerynumber($excellent); #get excellent+wrd count

		#Notes
		#n11 => reviewhorrible   reviewexcellent
		#n1+ => horriblecount    excellentcount
		#n+1 => reviewcount      reviewcount
		#n++ => totalcount       totalcount
		#m11 => badexpected      goodexpected

		#calculate the expected values
		$goodexpected = ($reviewcount * $excellentcount) / $totalcount;
		$badexpected = ($reviewcount * $horriblecount) / $totalcount;

		#calculate the dice coefficients and  cache results
		$dicegood += (2 * $reviewexcellent) / ($reviewcount + $excellentcount );
		$dicebad += (2 * $reviewhorrible) / ($reviewcount + $horriblecount );
		$goodDiceReviewCache{$word} = (2*$reviewexcellent)/($reviewcount+$excellentcount);
		$badDiceReviewCache{$word} = (2*$reviewhorrible)/($reviewcount+$horriblecount);

		#calculate the poisson measure and cache the results
		if($goodexpected != 0 && $reviewexcellent != 0 && $badexpected != 0 && $reviewhorrible != 0) {
		    $poissongood += ($reviewexcellent * ( log($reviewexcellent) - log($goodexpected) - 1) );
		    $goodPoissonReviewCache{$word} = ($reviewexcellent * ( log($reviewexcellent) - log($goodexpected) - 1) );
		    $poissonbad += ($reviewhorrible * ( log($reviewhorrible) - log($badexpected) - 1) );
		    $badPoissonReviewCache{$word} = ($reviewhorrible * ( log($reviewhorrible) - log($badexpected) - 1) );
		}

		#calculate the t-score and cache the results
		if($reviewexcellent != 0 && $reviewhorrible != 0 ) {
		    $tscoregood += ($reviewexcellent - $goodexpected)/sqrt($reviewexcellent);
		    $goodTscoreReviewCache{$word} = ($reviewexcellent - $goodexpected)/sqrt($reviewexcellent);
		    $tscorebad += ($reviewhorrible - $badexpected)/sqrt($reviewhorrible);
		    $badTscoreReviewCache{$word} = ($reviewhorrible - $badexpected)/sqrt($reviewhorrible);
		}
	    }
	}
    }

    #print results to trace
    print TRACE "DICE     GOOD: $dicegood    BAD: $dicebad  ";
    print TRACE "POISSON  GOOD: $poissongood BAD: $poissonbad ";
    print TRACE "T SCORE  GOOD: $tscoregood  BAD: $tscorebad ";


    my $results = 0;
    #calculate their agreement
    if($poissongood > $poissonbad) { $results += 1; } else { $results = $results - 1; }
    if($dicegood < $dicebad) { $results = $results - 1; } else { $results += 1; }
    if($tscoregood > $tscorebad) { $results += 1; } else { $results -= 1; }
    
    #return results
    ($results == 0) ? (return 0) : ($results < 0 ? (return -1) : (return 1) );
   
}

#determines the classification of the review using the synonyms from the Big
#Mac Machine Readable Dictionary
sub synonymTest{
    $totalscore=0;
    @tokens = split(/ /,$review);
    $negationflag=0;
    foreach $token (@tokens){
#	if($token=~m/(not|could|nothing|never|')/){ #'
#		      $negationflag=1;
#	}
#	   if($token=~m/(\.|\?|!)/){
#	       $negationflag=0;
#	   }
	$tokenhash{$token}=1;
    }
    foreach $token (keys %tokenhash){
#	if(exists $adj{$token}){
#	if($ldoce->isadjective($token)){
	    $sum=0;
	    # get synonyms of the word in a hash
#	    %syns = $bm->getSynonyms($token);
	    %syns = $bm->getAdjThesCode($token);
	    ####debug

	    foreach $key (keys %syns){
		@dots = split(/\./,$key);
		$dothash{$dots[0]}=1;
	    }
	    @thes = sort keys %dothash;
	    foreach $dot (keys %dothash){
		# see if the word is +ve or -ve based on the tagged classification
		$sum += $keytags{$dot};

	    }
	    # accumulate the total score by individual words
	    # whole is sum of parts
	    if($sum>0){
#		print TRACE "$token is +ve (@thes)\n ";
		$totalscore++;
	    }elsif($sum<0){
#		print TRACE "$token is -ve (@thes)\n ";
		$totalscore--;
	    }else{
		#print TRACE "$token is neutral \n ";
	    }
	    #empty the hash of synonyms
	    foreach $key (keys %syns){ delete $syns{$key}; }
	    foreach $key (keys %dothash){ delete $dothash{$key};}
#        }
    }

    print TRACE "$totalscore ";
    # foreach $key (keys %tokenhash){delete $tokenhash{$key};}
    if($totalscore>0){
	return 1;
    }elsif($totalscore<0){
	return -1;
    }elsif($totalscore==0){
	return 0;
    }
}

#determines the classification of the review based on the distance vectors
#Right now the algorithm uses only words which are one link away from 
#preselected good/bad words for classification of the reviews.
sub distance{
    @tokens = split(/ /,$review);
    $totalscore=0;
    foreach $token (@tokens){
	$tokenhash{$token}=1;
    }
		
    foreach $token (keys %tokenhash){
	if(exists $goodhash{$token}){        #check if token exists in list of good words
	    #print TRACE "$token is +ve\n";
	    $totalscore++;
	}elsif(exists $badhash{$token}){     #or in the list of bad words
	    #print TRACE "$token is -ve\n";
	    $totalscore--;                   #and adjust the totalscore accordingly
	}
    }
    foreach $key (keys %tokenhash){
	delete $tokenhash{$key};
    }
    if($totalscore>0){                       #make decision based on value of totalscore
	print TRACE "totalscore = $totalscore\n hence +ve review";
	return 1;
    }elsif($totalscore<0){
	print TRACE "totalscore = $totalscore\n hence -ve review";
	return -1;
    }elsif($totalscore==0){
	print TRACE "totalscore = $totalscore\n hence neutral review";
	return 0;
    }
}
